# -*- coding: utf-8 -*-
import json

from six import add_metaclass
from copy import deepcopy
from flask import current_app
from ldap3 import ObjectDef
from ldap3 import LDAPEntryError
from ldap3.utils.dn import safe_dn
from ldap3.utils.conv import check_json_dict, format_json

from .query import BaseQuery
from .attribute import LDAPAttribute


__all__ = ('LDAPEntry',)


class LDAPEntryMeta(type):

    # requiered
    base_dn = None
    entry_rdn = ['cn']
    object_classes = ['top']

    # optional
    sub_tree = True
    operational_attributes = False

    def __init__(cls, name, bases, ns):
        cls._attributes = dict()
        cls._object_def = ObjectDef(cls.object_classes)

        # loop through the namespace looking for LDAPAttribute instances
        for key, value in ns.items():
            if isinstance(value, LDAPAttribute):
                cls._attributes[key] = value
                attr_def = value.get_abstract_attr_def(key)
                cls._object_def.add(attr_def)

    @property
    def query(cls):
        return BaseQuery(cls)

    def get_new_type(cls):
        class_dict = deepcopy(cls()._attributes)
        new_cls = type(cls.__name__, (LDAPEntry,), class_dict)
        return new_cls


@add_metaclass(LDAPEntryMeta)
class LDAPEntry(object):

    def __init__(self, dn=None, changetype='add', **kwargs):
        self.__dict__['_dn'] = dn
        self.__dict__['_changetype'] = changetype

        for key, value in kwargs.items():
            if key not in self._attributes:
                raise LDAPEntryError('attribute not found')
            self._attributes[key]._init = value

    def __iter__(self):
        for attribute in self._attributes:
            yield self._attributes[attribute]

    def __contains__(self, item):
        return True if self.__getitem__(item) else False

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __getattr__(self, item):
        if item not in self._attributes:
            return None

        return self._attributes[item]

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __setattr__(self, key, value):
        if key not in self._attributes:
            raise LDAPEntryError('attribute not found')

        self._attributes[key].value = value

    @property
    def dn(self):
        if self._dn is None:
            self.generate_dn_from_entry()
        return self._dn

    def generate_dn_from_entry(self):
        rdn_list = list()
        for attr in self._object_def:
            if attr.name in self.entry_rdn:
                if len(self._attributes[attr.key]) == 1:
                    rdn = '{attr}={value}'.format(
                        attr=attr.name,
                        value=self._attributes[attr.key].value
                    )
                    rdn_list.append(rdn)

        dn = '{rdn},{base_dn}'.format(rdn='+'.join(rdn_list),
                                      base_dn=self.base_dn)

        self.__dict__['_dn'] = safe_dn(dn)

    def get_attributes_dict(self):
        return dict((attribute_key, attribute_value.values) for (attribute_key,
                    attribute_value) in self._attributes.items())

    def get_entry_add_dict(self, attr_dict):
        add_dict = dict()
        for attribute_key, attribute_value in attr_dict.items():
            if self._attributes[attribute_key].value:
                attribute_def = self._object_def[attribute_key]
                add_dict.update({attribute_def.name: attribute_value})
        return add_dict

    def get_entry_modify_dict(self, attr_dict):
        modify_dict = dict()
        for attribute_key, attribute_value in attr_dict.items():
            if self._attributes[attribute_key].changetype is not None:
                attribute_def = self._object_def[attribute_key]
                changes = self._attributes[attribute_key].get_changes_tuple()
                modify_dict.update({attribute_def.name: changes})
        return modify_dict

    @property
    def connection(self):
        app = current_app._get_current_object()
        ldapc = app.extensions.get('ldap_conn')
        return ldapc

    def delete(self):
        '''Delete this entry from LDAP server'''
        self.connection.connection.delete(self.dn)

    def save(self):
        '''Save the current instance'''
        attributes = self.get_entry_add_dict(self.get_attributes_dict())
        if self._changetype is 'add':
            return self.connection.connection.add(self.dn,
                                                  self.object_classes,
                                                  attributes)
        elif self._changetype == 'modify':
            changes = self.get_entry_modify_dict(self.get_attributes_dict())
            return self.connection.connection.modify(self.dn, changes)

        return False

    def authenticate(self, password):
        '''Authenticate a user with an LDAPModel class

        Args:
            password (str): The user password.

        '''
        return self.connection.authenticate(self.dn, password)

    def to_json(self, indent=2, sort=True):
        json_entry = dict()
        json_entry['dn'] = self.dn
        json_entry['attributes'] = self.get_attributes_dict()

        if str == bytes:
            check_json_dict(json_entry)

        json_output = json.dumps(json_entry,
                                 ensure_ascii=True,
                                 sort_keys=sort,
                                 indent=indent,
                                 check_circular=True,
                                 default=format_json,
                                 separators=(',', ': '))

        return json_output


LDAPModel = LDAPEntry



# -*- coding: utf-8 -*-
from ldap3 import AttrDef, LDAPAttributeError
from ldap3 import STRING_TYPES, MODIFY_ADD, MODIFY_DELETE, MODIFY_REPLACE


class LDAPAttribute(object):

    def __init__(self, name, primary=False, validate=None, default=None,
                 dereference_dn=None):
        self.__dict__['name'] = name
        self.__dict__['values'] = []
        self.__dict__['default'] = default
        self.__dict__['validate'] = validate
        self.__dict__['changetype'] = None
        self.__dict__['dereference_dn'] = dereference_dn

    def __str__(self):
        return self.value

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return self.values.__iter__()

    def __contains__(self, item):
        return item in self.__dict__['values']

    def __setattr__(self, item, value):
        if item not in ['value', '_init']:
            raise LDAPAttributeError('can not set key')

        # set changetype
        if item is 'value':
            if self.__dict__['values']:
                if not value:
                    self.__dict__['changetype'] = MODIFY_DELETE
                else:
                    self.__dict__['changetype'] = MODIFY_REPLACE
            else:
                self.__dict__['changetype'] = MODIFY_ADD

        if isinstance(value, STRING_TYPES):
            value = [value]
        self.__dict__['values'] = value

    @property
    def value(self):
        '''Return the single value or a list of values of the attribute.'''
        if len(self.__dict__['values']) == 1:
            return self.__dict__['values'][0]
        else:
            return self.__dict__['values']

    @property
    def changetype(self):
        return self.__dict__['changetype']

    def get_changes_tuple(self):
        values = [val.encode('UTF-8') for val in self.__dict__['values']]
        return (self.changetype, values)

    def append(self, value):
        '''Add another value to the attribute'''
        if self.__dict__['values']:
            self.__dict__['changetype'] = MODIFY_REPLACE

        self.__dict__['values'].append(value)

    def delete(self):
        '''Delete this attribute

        This property sets the value to an empty list an the changetype
        to delete.
        '''
        self.value = []

    def get_abstract_attr_def(self, key):
        return AttrDef(name=self.name, key=key,
                       validate=self.validate,
                       default=self.default,
                       dereference_dn=self.dereference_dn)



# -*- coding: utf-8 -*-
from flask import current_app
from ldap3 import Reader, utils


__all__ = ('BaseQuery',)


class BaseQuery(object):

    def __init__(self, type):
        self.type = type

        self.query = []
        self.base_dn = getattr(type, 'base_dn')
        self.sub_tree = getattr(type, 'sub_tree')
        self.object_def = getattr(type, '_object_def')
        self.operational_attributes = getattr(type, 'operational_attributes')

    def __iter__(self):
        for entry in self.get_reader_result():
            new_cls = self.type.get_new_type()
            ldapentry = new_cls(dn=entry.entry_get_dn(),
                                changetype='modify',
                                **entry.entry_get_attributes_dict())
            yield ldapentry

    def get_reader_result(self):
        query = ','.join(self.query)
        app = current_app._get_current_object()
        ldapc = app.extensions.get('ldap_conn')
        reader = Reader(connection=ldapc.connection,
                        object_def=self.object_def,
                        query=query,
                        base=self.base_dn,
                        components_in_and=True,
                        sub_tree=self.sub_tree,
                        get_operational_attributes=self.operational_attributes,
                        controls=None)
        reader.search()
        return reader.entries

    def get(self, ldap_dn):
        '''Return an LDAP entry

        Args:
            ldap_dn (str): LDAP DN
        '''
        rdns = utils.dn.to_dn(ldap_dn)
        user_rdn = rdns.pop(0)
        self.query.append('(' + user_rdn + ')')
        self.base_dn = ','.join(rdns)
        return self.first()

    def filter(self, *query_filter):
        '''Set the query filter to perform the query with

        Args:
            *query_filter: Simplified Query Language filter
        '''
        for query in query_filter:
            self.query.append(query)
        return self

    def first(self):
        '''Execute the query and return the first result

        If there are no entries, first returns ``None``
        '''
        for entry in iter(self):
            return entry
        return None

    def all(self):
        '''Return all of the results of a query in a list'''
        return [obj for obj in iter(self)]



# -*- coding: utf-8 -*-

'''
Flask-LDAPConn
--------------

Flask extension providing ldap3 connection object and ORM
to accessing LDAP servers.
'''


from setuptools import setup


setup(
    name='Flask-LDAPConn',
    version='0.6',
    url='http://github.com/rroemhild/flask-ldapconn',
    license='BSD',
    author='Rafael Römhild',
    author_email='rafael@roemhild.de',
    keywords='flask ldap ldap3 orm',
    description='Pure python, LDAP connection and ORM for Flask Applications',
    long_description=open('README.rst').read(),
    packages=[
        'flask_ldapconn'
    ],
    zip_safe=False,
    platforms='any',
    install_requires=[
        'Flask==0.10.1',
        'ldap3==0.9.7.10',
        'six==1.9.0'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Framework :: Flask',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# -*- coding: utf-8 -*-
from ssl import CERT_OPTIONAL, PROTOCOL_TLSv1
from flask import current_app
from flask import _app_ctx_stack as stack
from ldap3 import Server, Connection, Tls
from ldap3 import STRATEGY_SYNC, GET_ALL_INFO, SUBTREE
from ldap3 import AUTO_BIND_NO_TLS, AUTO_BIND_TLS_BEFORE_BIND
from ldap3 import LDAPBindError, LDAPInvalidFilterError, LDAPInvalidDnError
from ldap3.utils.dn import split_ava

from .entry import LDAPEntry
from .attribute import LDAPAttribute


__all__ = ('LDAPConn',)


class LDAPConn(object):

    def __init__(self, app=None):

        self.Entry = LDAPEntry
        self.Attribute = LDAPAttribute

        self.Model = self.Entry

        if app is not None:
            self.init_app(app)

    def init_app(self, app):

        # Default config
        app.config.setdefault('LDAP_SERVER', 'localhost')
        app.config.setdefault('LDAP_PORT', 389)
        app.config.setdefault('LDAP_BINDDN', None)
        app.config.setdefault('LDAP_SECRET', None)
        app.config.setdefault('LDAP_TIMEOUT', 10)
        app.config.setdefault('LDAP_USE_SSL', False)
        app.config.setdefault('LDAP_USE_TLS', True)
        app.config.setdefault('LDAP_REQUIRE_CERT', CERT_OPTIONAL)
        app.config.setdefault('LDAP_CERT_PATH', None)

        self.tls = Tls(validate=app.config['LDAP_REQUIRE_CERT'],
                       version=PROTOCOL_TLSv1,
                       ca_certs_file=app.config['LDAP_CERT_PATH'])

        self.ldap_server = Server(
            host=app.config['LDAP_SERVER'],
            port=app.config['LDAP_PORT'],
            use_ssl=app.config['LDAP_USE_SSL'],
            get_info=GET_ALL_INFO
        )

        # Store ldap_conn object to extensions
        app.extensions['ldap_conn'] = self

        # Teardown appcontext
        app.teardown_appcontext(self.teardown)

    def connect(self, user, password):
        auto_bind_strategy = AUTO_BIND_TLS_BEFORE_BIND
        if current_app.config['LDAP_USE_TLS'] is not True:
            auto_bind_strategy = AUTO_BIND_NO_TLS

        ldap_conn = Connection(
            self.ldap_server,
            auto_bind=auto_bind_strategy,
            client_strategy=STRATEGY_SYNC,
            user=user,
            password=password,
            check_names=True
        )

        return ldap_conn

    def teardown(self, exception):
        ctx = stack.top
        if hasattr(ctx, 'ldap_conn'):
            ctx.ldap_conn.unbind()

    @property
    def connection(self):
        user = current_app.config['LDAP_BINDDN']
        password = current_app.config['LDAP_SECRET']

        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'ldap_conn'):
                ctx.ldap_conn = self.connect(user, password)
            return ctx.ldap_conn

    def authenticate(self,
                     username,
                     password,
                     attribute=None,
                     base_dn=None,
                     search_filter=None,
                     search_scope=SUBTREE):
        '''Attempts to bind a user to the LDAP server.

        Args:
            username (str): DN or the username to attempt to bind with.
            password (str): The password of the username.
            attribute (str): The LDAP attribute for the username.
            base_dn (str): The LDAP basedn to search on.
            search_filter (str): LDAP searchfilter to attempt the user
                search with.

        Returns:
            bool: ``True`` if successful or ``False`` if the
                credentials are invalid.
        '''
        # If the username is no valid DN we can bind with, we nee to find
        # the user first.
        if not split_ava(username)[0]:
            user_filter = '({0}={1})'.format(attribute, username)
            if search_filter is not None:
                user_filter = '(&{0}{1})'.format(user_filter, search_filter)

            try:
                self.connection.search(base_dn, user_filter, search_scope,
                                       attributes=[attribute])
                response = self.connection.response
                username = response[0]['dn']
            except (LDAPInvalidDnError, LDAPInvalidFilterError, IndexError):
                return False

        try:
            conn = self.connect(username, password)
            conn.unbind()
            return True
        except LDAPBindError:
            return False

    def whoami(self):
        '''Deprecated

        Use LDAPConn.connection.extend.standard.who_am_i()
        '''
        return self.connection.extend.standard.who_am_i()

    def result(self):
        '''Deprecated

        Use LDAPConn.connection.result
        '''
        return self.connection.result

    def response(self):
        '''Deprecated

        Use LDAPConn.connection.response
        '''
        return self.connection.response

    def search(self, *args, **kwargs):
        '''Deprecated

        Use LDAPConn.connection.search()
        '''
        return self.connection.search(*args, **kwargs)
